
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include "log.h"
#include "struct.h"
#include "irc_printf.h"

extern int errno;

typedef struct mem_struct MemList;

#define GC_MAGIC 0x3bd7fb89

struct mem_struct {
     unsigned int magic;
     void *ptr;
     int size;
     int refcount;
     pthread_mutex_t mutex;
     MemList *prev, *next;
};

struct m_ptr_struct {
     int magic;
};

volatile static unsigned int is_initialized = 0;

MemList *memList = NULL;

static pthread_t thr;

void gcCancel()
{
     pthread_cancel(thr);
}

void *__gcThread(void *);

/* __gcInit()
   initializes the garbage collector.
 */

__gcInit()
{
     pthread_t th;

     log("initializing garbage collector..");
     memList = malloc(sizeof(MemList));
     memList->magic = GC_MAGIC;
     memList->ptr = NULL;
     memList->size = 0;
     memList->refcount = 1;	/* refcount must ALWAYS be 1 for this entry, otherwise it
				   will be freed by the gc thread */
     memList->prev = memList->next = NULL;
     pthread_mutex_init(&(memList->mutex), NULL);
     pthread_create(&th, NULL, __gcThread, NULL);
     thr = th;
     is_initialized = 1;
}

void gc_implicit_init()
{
     if (!is_initialized) {
	  __gcInit();
     }
     return;
}

/* gcAlloc(size)
   returns a segment of memory which will be automatically garbage collected when no
   longer in use.
 */

void *gcAlloc(size_t size)
{
     char *myptr = malloc(size);
     MemList *memhdr = malloc(sizeof(MemList));
     char *r_memptr = myptr;

     gc_implicit_init();

     log("entering gcAlloc..");
     bzero(memhdr, sizeof(MemList));
     memhdr->magic = GC_MAGIC;
     memhdr->ptr = r_memptr;
     memhdr->size = size;
     memhdr->refcount = 1;
     pthread_mutex_init(&(memhdr->mutex), NULL);
     bzero(r_memptr, size);
     log("initialized structure and memory, adding to block list");
     {
	  MemList *mem_l = memList;

	  if (!mem_l)
	       __gcInit();
	  while (mem_l->next)
	       mem_l = mem_l->next;
	  mem_l->next = memhdr;
	  memhdr->prev = mem_l;
     }
     log("returning.");
     return (void *) r_memptr;
}


/* gcFree does not exist - that is what garbage collecting is for.
   functions are needed though to increase the refcount on a particular block
   of memory, and to decrease it when it is no longer in use.
 */

/* int gcChgRefcnt(void *ptr, int chg)
   changes the reference count for block ptr by chg
 */

int gcChgRefcnt(void *ptr, int chg)
{
     MemList *mem_l = memList;
     MemList *memhdr;
     char *myptr = ptr;
     int i;

     gc_implicit_init();
     mem_l = memList;

     while (mem_l) {
	  if (mem_l->ptr == ptr) {
	       memhdr = mem_l;
	       mem_l = NULL;
	  }
	  if (mem_l)
	       mem_l = mem_l->next;
     }

     if (memhdr) {
	  log("entering critical section..");
	  pthread_mutex_lock(&(memhdr->mutex));
	  if (memhdr->magic != GC_MAGIC) {
	       char buf[255];

	       snprintf(buf, 255, "gcChgRefcnt(%p): invalid magic for block!", ptr);
	       log(buf);
	  } {
	       char buf[255];

	       snprintf(buf, 255, "gcChgRefcnt(%p) called, refcnt for %p is %i, new refcnt is %i",
			ptr, ptr, memhdr->refcount, memhdr->refcount + chg);
	       log(buf);
	  }
	  memhdr->refcount += chg;
	  pthread_mutex_unlock(&(memhdr->mutex));
	  log("exiting critical section..");
	  return memhdr->refcount;
     } else {
	  errno = EINVAL;
	  return -1;
     }
}

/* gcIncRefcnt - tells the garbage collector that this structure is being used in another
   place.
 */

int gcIncRefcnt(void *ptr)
{
     return gcChgRefcnt(ptr, 1);
}

/* gcDecRefcnt - tells the garbage collector that this structure is no longer in use by any
   particular function.
 */

int gcDecRefcnt(void *ptr)
{
     return gcChgRefcnt(ptr, -1);
}

/* gcDoEnterExitCriticalSection - enters or exits a critical section on a portion of memory.
   if type == 1, locks the mutex
   if type == 2, tries to lock the mutex (does not block)
   if type == 3, unlocks the mutex
   otherwise returns -1 and sets errno to EINVAL
 */

int gcDoEnterExitCriticalSection(void *ptr, int type)
{
     MemList *q;

     gc_implicit_init();

     q = memList->next;
     while (q) {
	  if (q->ptr == ptr) {
	       if (type == 1) {
		    return pthread_mutex_lock(&(q->mutex));
	       } else if (type == 2) {
		    return pthread_mutex_trylock(&(q->mutex));
	       } else if (type == 3) {
		    return pthread_mutex_unlock(&(q->mutex));
	       } else {
		    errno = EINVAL;
		    return -1;
	       }
	  }
	  q = q->next;
     }

     errno = EINVAL;
     return -1;
}

/* gcEnterCriticalSection - enters a critical section, and locks the memory associated with the ptr */

int gcEnterCriticalSection(void *ptr)
{
     return gcDoEnterExitCriticalSection(ptr, 1);
}

/* gcTryEnterCriticalSection - attempts to enter a critical section, returning whether or not 
   the ptr was locked successfully
 */
int gcTryEnterCriticalSection(void *ptr)
{
     return gcDoEnterExitCriticalSection(ptr, 2);
}

/* gcExitCriticalSection - exits a critical section by unlocking the memory */
int gcExitCriticalSection(void *ptr)
{
     return gcDoEnterExitCriticalSection(ptr, 3);
}

#ifndef TIME_TO_GC
#define TIME_TO_GC 60
#endif

/* __gcThread()
   this is the thread that runs every 60 seconds (or at another interval if TIME_TO_GC is set)
   to collect garbage.
 */

void *__gcThread(void *arg)
{
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
     while (1) {
	  log("entering __gcThread");
	  sleep(TIME_TO_GC);
	  {
	       MemList *l = memList->next;

	       log("beginning garbage collection loop");
	       while (l) {
		    if (l->refcount == 0) {
			 log("entering critical section..");
			 if (!pthread_mutex_trylock(&(l->mutex))) {
			      {
				   char buf[255];

				   snprintf(buf, 255, "refcnt for %p is 0, freeing block", l->ptr);
				   log(buf);
			      }
			      if (l->next) {
				   MemList *p = l->prev;

				   (l->prev)->next = l->next;
				   (l->next)->prev = l->prev;
				   free(l->ptr);
				   log("exiting critical section..");
				   pthread_mutex_unlock(&(l->mutex));
				   free(l);
				   l = p;
			      } else {
				   MemList *p = l->prev;

				   (l->prev)->next = NULL;
				   free(l->ptr);
				   log("exiting critical section..");
				   pthread_mutex_unlock(&(l->mutex));
				   free(l);
				   l = p;
			      }
			 } else {
			      log("could not lock mutex, skipping this block..");
			 }
		    }
		    l = l->next;
	       }
	  }
     }
}

/* m_meminfo
   av[0] = "MEMINFO"
   returns a listing of allocated blocks, and the types of the blocks
 */

#define fprintf irc_fprintf

void m_meminfo(struct fd_list *fd, char **av, int ac)
{
     struct m_ptr_struct *r;
     MemList *q;

     gc_implicit_init();
     q = memList->next;
     while (q) {
	  if (q->magic == GC_MAGIC) {
	       char *block_type;

	       r = q->ptr;
	       switch (r->magic) {
	       case FD_MAGIC:
		    block_type = "fd_list block";
		    break;
	       case SERVER_MAGIC:
		    block_type = "server_list block";
		    break;
	       case USER_LIST_MAGIC:
		    block_type = "user_list block";
		    break;
	       case INVITE_LIST_MAGIC:
		    block_type = "invite_list block";
		    break;
	       case BAN_LIST_MAGIC:
		    block_type = "ban_list block";
		    break;
	       case CHAN_LIST_MAGIC:
		    block_type = "chan_list block";
		    break;
	       case USER_CHAN_LIST_MAGIC:
		    block_type = "user_chan_list block";
		    break;
	       case USER_INVITE_LIST_MAGIC:
		    block_type = "user_invite_list block";
		    break;
	       case OLINE_MAGIC:
		    block_type = "oline block";
		    break;
	       case PLINE_MAGIC:
		    block_type = "pline block";
		    break;
	       case YLINE_MAGIC:
		    block_type = "yline block";
		    break;
	       case ILINE_MAGIC:
		    block_type = "iline block";
		    break;
	       case MOTD_MAGIC:
		    block_type = "motd block";
		    break;
	       case WHOWAS_MAGIC:
		    block_type = "whowas block";
		    break;
	       default:
		    block_type = "(void *) block";
	       }

	       fprintf(fd->sok, ":%s NOTICE %s :Block %p: size=%i, refcount=%i, type=%s\r\n",
		       server_name, fd->nick, q->ptr, q->size, q->refcount, block_type);
	  } else {
	       fprintf(fd->sok, ":%s NOTICE %s :Block %p: BAD MAGIC!!\r\n",
		       server_name, fd->nick, q->ptr);
	  }
	  q = q->next;
     }
}
