#include <stdarg.h>
#include <stdio.h>
#include "struct.h"
#include <string.h>
#include <errno.h>
#include "gc.h"
#include "irc_printf.h"

extern int errno;

#define GLINE_MAGIC 0xdeadbeef

struct gline {
     int magic;
     struct gline *prev, *next;
     char hostmask[100];
     char reason[100];
     int expire_time;
     int do_delete;
     int is_enabled;
};

struct gline *glist = NULL;

/* __init_glines()
   initializes the gline list
 */

__init_glines()
{
     glist = (struct gline *) gcAlloc(sizeof(struct gline));

     if (!glist) {
	  fprintf(stderr, "can't allocate gline list!\n");
	  exit(-1);
     }
     bzero(glist, sizeof(struct gline));

     glist->magic = GLINE_MAGIC;
     return;
}

char upper[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ{}|";

/* void strtolower(char *i)
   does an in-place conversion of the string in i so that it is all lowercase
 */

void strtolower(char *i)
{
     int j;

     for (j = 0; j < strlen(i); j++) {
	  if (strchr(upper, i[j])) {
	       i[j] += ' ';
	  }
     }
}

/* int match(const char *hm, const char *hn)
   attempts to perform shell-style globbing on hm to match it to hn
 */

int match(const char *_hm, const char *_hn)
{
     char hn[strlen(_hn) + 1], hm[strlen(_hm) + 1];

     bzero(hn, strlen(_hn) + 1);
     bzero(hm, strlen(_hm) + 1);
     bcopy(_hn, hn, strlen(_hn) + 1);
     bcopy(_hm, hm, strlen(_hm) + 1);

     if (strlen(hm) == 1 && hm[0] == '*') {	/* anything matches this one */
	  return 0;
     }
     if (!strchr(hm, '*') && !strchr(hm, '?') && !strcasecmp(hm, hn)) {		/* no * or ? and they match */
	  return 0;
     } else if (!strchr(hm, '*') && !strchr(hm, '?') && strcasecmp(hm, hn)) {
	  return 1;
     }
     if (!strchr(hm, '*') && strchr(hm, '?')) {
	  int i;

	  if (strlen(hm) != strlen(hn)) {
	       return 1;
	  }
	  strtolower(hm);
	  strtolower(hn);

	  i = 0;
	  for (i = 0; i < strlen(hm); i++) {
	       if (hm[i] != '?') {
		    if (hm[i] != hn[i]) {
			 return 1;
		    }
	       }
	  }
	  return 0;
     } else {
	  int i, j, star;

	  strtolower(hm);
	  strtolower(hn);

	  j = 0;
	  for (i = 0; i < strlen(hn); i++) {
	       if (hm[j] == '?') {
		    j++;
	       } else if (hm[j] == '*') {
		    star = 1;
		    j++;
		    if (hm[j] == hn[i] && (hm[j + 1] == hn[i + 1] || hm[j + 1] == '*' || hm[j + 1] == '?')) {
			 star = 0;
			 j++;
		    }
	       } else {
		    if (hm[j] == hn[i] && (hm[j + 1] == hn[i + 1] || hm[j + 1] == '*' || hm[j + 1] == '?')) {
			 star = 0;
			 j++;
		    } else {
			 if (!star) {
			      return 1;
			 }
		    }
	       }
	  }
	  if (strlen(hm) > j) {
	       return 1;
	  } else {
	       return 0;
	  }
     }

}

/* find_gline
 * returns a pointer to a gline structure or NULL if not found
 *
 * returns NULL and errno EINVAL if hm is NULL
 */

struct gline *find_gline(char *hm, int exact)
{
     struct gline *q = glist;
     int (*compare) (const char *, const char *);

     if (exact)
	  compare = strcasecmp;
     else
	  compare = match;

     if (!hm) {
	  errno = EINVAL;
	  return NULL;
     }
     while (q) {
	  if (!compare(q->hostmask, hm)) {
	       return q;
	  }
	  q = q->next;
     }
     errno = 0;
     return NULL;
}

/* construct_gline
 * returns a filled 'struct gline'
 *
 * if unable to allocate memory for gline, returns NULL with errno ENOMEM
 * if hm is invalid, returns NULL with errno EINVAL
 * else returns filled structure
 */

struct gline *construct_gline(char *hm, char *reason, int seconds_to_expire)
{
     struct gline *tmp = (struct gline *) gcAlloc(sizeof(struct gline));
     int at, pos;

     if (!tmp) {
	  errno = ENOMEM;
	  return NULL;
     }
     bzero(tmp, sizeof(struct gline));

     snprintf(tmp->hostmask, 100, "%s", hm);

     if (reason) {
	  strncpy(tmp->reason, reason, 100);
     } else {
	  snprintf(tmp->reason, 100, "G-lined");
     }

     if (seconds_to_expire) {
	  tmp->expire_time = __time(NULL) + seconds_to_expire;
     } else {
	  tmp->expire_time = __time(NULL) + 3600;	/* 1 hour */
     }

     tmp->magic = GLINE_MAGIC;
     return tmp;
}

/* link_gline
 * adds a gline to the list of glines
 * 
 * returns -1 with EINVAL if q is not a valid gline structure
 * else returns 0
 */

int link_gline(struct gline *q)
{
     struct gline *p = glist;

     if (!q || q->magic != GLINE_MAGIC) {
	  errno = EINVAL;
	  return -1;
     }
     q->next = p;
     if (p) {
	  p->prev = q;
     }
     glist = q;
}

/* expire_glines
 * removes expired glines from the list
 *
 */

int expire_glines()
{
     struct gline *p = glist;
     int now = __time(NULL);

     while (p) {
	  if (p->expire_time < now) {
	       p->do_delete = 1;
	  }
	  p = p->next;
     }

     p = glist;
     while (p) {
	  if (p->do_delete) {
	       if (p == glist) {	/* head of list */
		    if (!p->next) {
			 p->magic = 0;
			 gcDecRefcnt(p);
			 glist = NULL;
		    } else {
			 (p->next)->prev = NULL;	/* 138 */
			 glist = p->next;	/* 139 */

			 p->magic = 0;	/* 141 */
			 gcDecRefcnt(p);	/* 142 */

			 p = glist;	/* 144 */
		    }
	       } else if (p->prev != NULL && p->next != NULL) {		/* 145 */
		    struct gline *q = p->next;	/* 146 */

		    (p->prev)->next = p->next;	/* 147 */
		    (p->next)->prev = p->prev;	/* 148 */

		    p->magic = 0;	/* 150 */
		    gcDecRefcnt(p);	/* 151 */

		    p = q;	/* 153 */
	       } else if (p->prev != NULL && p->next == NULL) {
		    (p->prev)->next = NULL;

		    p->magic = 0;
		    gcDecRefcnt(p);
		    p = NULL;
	       }
	  }
	  if (p)
	       p = p->next;
     }

     return 0;
}

/* void broadcast_to_ops(char *fmt, ...)
   sends a message to all operators consisting of the format string in fmt and all of the args after it
 */

#define vfprintf irc_vfprintf
#define fprintf irc_fprintf

void broadcast_to_ops(char *fmt,...)
{
     va_list ap;
     char buf[255];
     struct fd_list *q = __head->next;

     va_start(ap, fmt);
     while (q) {
	  if (q->oper) {
	       snprintf(buf, 255, ":%s NOTICE %s :%s\r\n", server_name, q->nick, fmt);
	       vfprintf(q->sok, buf, ap);
	  }
	  q = q->next;
     }
     va_end(ap);
}

/* m_gline
 * the irc command for adding a gline
 *
 * av[0] = command
 * av[1] = gline mask
 * av[2] = length of gline
 * av[3] = comment
 */

void m_gline(struct fd_list *fd, char **av, int ac)
{
     if (ac == 1) {
	  char *greply = ":%s 280 %s %s %i %s\r\n";
	  char *gendreply = ":%s 281 %s :End of GLINE list.\r\n";
	  struct gline *q = glist;

	  while (q) {
	       fprintf(fd->sok, greply, server_name, fd->nick, q->hostmask, q->expire_time, q->reason);
	       q = q->next;
	  }
	  fprintf(fd->sok, gendreply, server_name, fd->nick);
	  return;
     } else if (ac == 2) {
	  if (!strcasecmp(av[0], "STATS")) {	/* we're being called for a stats g reply */
	       char *greply = ":%s 247 %s %c %s %i :%s\r\n";
	       struct gline *q = glist;

	       while (q) {
		    fprintf(fd->sok, greply, server_name, fd->nick, 'G', q->hostmask, q->expire_time, q->reason);
		    q = q->next;
	       }
	       return;
	  } else {
	       char *greply = ":%s 280 %s %s %i %s\r\n";
	       char *gendreply = ":%s 281 %s :End of GLINE list.\r\n";
	       struct gline *q = glist;

	       if (fd->oper) {
		    if (av[1][0] == '+' || av[1][0] == '-') {
			 if (av[1][0] == '+') {
			      struct gline *q = glist;

			      while (q) {
				   if (!strcasecmp(q->hostmask, &(av[1][1]))) {
					q->is_enabled = 1;
				   }
				   q = q->next;
			      }
			 } else if (av[1][0] == '-') {
			      struct gline *q = glist;

			      while (q) {
				   if (!strcasecmp(q->hostmask, &(av[1][1]))) {
					q->is_enabled = 0;
				   }
				   q = q->next;
			      }
			 }
			 return;
		    }
	       }
	       if (av[1][0] == 0) {
		    while (q) {
			 fprintf(fd->sok, greply, server_name, fd->nick, q->hostmask, q->expire_time, q->reason);
			 q = q->next;
		    }
	       } else {
		    while (q) {
			 if (!match(q->hostmask, av[1])) {
			      fprintf(fd->sok, greply, server_name, fd->nick, q->hostmask, q->expire_time, q->reason);
			 }
			 q = q->next;
		    }
	       }
	       fprintf(fd->sok, gendreply, server_name, fd->nick);
	       return;
	  }
     } else if (ac == 3) {
	  if (!strcasecmp(av[0], "STATS")) {
	       ac--;
	       m_gline(fd, av, ac);
	       return;
	  } else {
	       __not_enough_parameters(fd, av[0]);
	       return;
	  }
     } else if (ac >= 4) {
	  if (!fd->oper) {
	       __no_priveliges(fd);
	       return;
	  } else {
	       int myptr = 0;
	       struct gline *q;

	       /* av[0] = cmd, av[1] = hm, av[2] = exp time, av[3] = reason.. */
	       /* is av[1][0] == '+' ? */
	       if (av[1][0] == '+')
		    myptr = 1;

	       q = find_gline(&(av[1][myptr]), 1);
	       if (q) {
		    broadcast_to_ops("%s resetting expiration time on GLINE for %s to %i", fd->nick, q->hostmask, __time(NULL) + atoi(av[2]));
		    q->expire_time = __time(NULL) + atoi(av[2]);
	       } else {
		    broadcast_to_ops("%s adding local GLINE for %s, expiring at %i: %s", fd->nick, &(av[1][myptr]), __time(NULL) + atoi(av[2]),
				     av[3]);
		    q = construct_gline(&(av[1][myptr]), av[3], atoi(av[2]));
		    link_gline(q);
	       }
	       return;
	  }
     }
}
